#   Copyright 2017 GoDaddy
#   Copyright 2019 Red Hat, Inc. All rights reserved.
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Pool action implementation"""

from cliff import lister
from osc_lib.command import command
from osc_lib import exceptions
from osc_lib import utils
from osc_lib.utils import tags as _tag
from oslo_utils import uuidutils

from octaviaclient.osc.v2 import constants as const
from octaviaclient.osc.v2 import utils as v2_utils

PROTOCOL_CHOICES = ['TCP', 'HTTP', 'HTTPS', 'PROXY', 'PROXYV2', 'UDP',
                    'SCTP']
ALGORITHM_CHOICES = ['SOURCE_IP', 'ROUND_ROBIN', 'LEAST_CONNECTIONS',
                     'SOURCE_IP_PORT']


class CreatePool(command.ShowOne):
    """Create a pool"""

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)

        parser.add_argument(
            '--name',
            metavar='<name>',
            help="Set pool name."
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help="Set pool description."
        )
        parser.add_argument(
            '--protocol',
            metavar='{' + ','.join(PROTOCOL_CHOICES) + '}',
            required=True,
            choices=PROTOCOL_CHOICES,
            type=lambda s: s.upper(),  # case insensitive
            help="Set the pool protocol."
        )
        parent_group = parser.add_mutually_exclusive_group(required=True)
        parent_group.add_argument(
            '--listener',
            metavar='<listener>',
            help="Listener to add the pool to (name or ID)."
        )
        parent_group.add_argument(
            '--loadbalancer',
            metavar='<load_balancer>',
            help="Load balancer to add the pool to (name or ID)."
        )
        parser.add_argument(
            '--session-persistence',
            metavar='<session persistence>',
            help="Set the session persistence for the listener (key=value)."
        )
        parser.add_argument(
            '--lb-algorithm',
            metavar='{' + ','.join(ALGORITHM_CHOICES) + '}',
            required=True,
            choices=ALGORITHM_CHOICES,
            type=lambda s: s.upper(),  # case insensitive
            help="Load balancing algorithm to use."
        )
        admin_group = parser.add_mutually_exclusive_group()
        admin_group.add_argument(
            '--enable',
            action='store_true',
            default=True,
            help="Enable pool (default)."
        )
        admin_group.add_argument(
            '--disable',
            action='store_true',
            default=None,
            help="Disable pool."
        )
        parser.add_argument(
            '--tls-container-ref',
            metavar='<container-ref>',
            help="The reference to the key manager service secrets container "
                 "containing the certificate and key for ``tls_enabled`` "
                 "pools to re-encrpt the traffic to backend member servers."
        )
        parser.add_argument(
            '--ca-tls-container-ref',
            metavar='<ca_tls_container_ref>',
            help="The reference to the key manager service secrets container "
                 "containing the CA certificate for ``tls_enabled`` pools "
                 "to check the backend member servers certificates."
        )
        parser.add_argument(
            '--crl-container-ref',
            metavar='<crl_container_ref>',
            help="The reference to the key manager service secrets container "
                 "containting the CA revocation list file for ``tls_enabled`` "
                 "pools to validate the backend member servers certificates."
        )
        tls_enable = parser.add_mutually_exclusive_group()
        tls_enable.add_argument(
            '--enable-tls',
            action='store_true',
            default=None,
            help="Enable backend member re-encryption."
        )
        tls_enable.add_argument(
            '--disable-tls',
            action='store_true',
            default=None,
            help="Disable backend member re-encryption."
        )
        parser.add_argument(
            '--wait',
            action='store_true',
            help='Wait for action to complete.',
        )
        parser.add_argument(
            '--tls-ciphers',
            metavar='<tls_ciphers>',
            help="Set the TLS ciphers to be used by the pool "
                 "in OpenSSL cipher string format."
        )
        parser.add_argument(
            '--tls-version',
            dest='tls_versions',
            metavar='<tls_versions>',
            nargs='?',
            action='append',
            help="Set the TLS protocol version to be used "
                 "by the pool (can be set multiple times)."
        )
        parser.add_argument(
            '--alpn-protocol',
            dest='alpn_protocols',
            metavar='<alpn_protocols>',
            nargs='?',
            action='append',
            help="Set the ALPN protocol to be used "
                 "by the pool (can be set multiple times)."
        )

        _tag.add_tag_option_to_parser_for_create(
            parser, 'pool')

        return parser

    def take_action(self, parsed_args):
        rows = const.POOL_ROWS
        attrs = v2_utils.get_pool_attrs(self.app.client_manager, parsed_args)

        body = {"pool": attrs}
        data = self.app.client_manager.load_balancer.pool_create(
            json=body)

        if parsed_args.wait:
            v2_utils.wait_for_active(
                status_f=(self.app.client_manager.load_balancer.
                          load_balancer_show),
                res_id=data['pool']['loadbalancers'][0]['id']
            )
            data = {
                'pool': (
                    self.app.client_manager.load_balancer.pool_show(
                        data['pool']['id']))
            }

        formatters = {'loadbalancers': v2_utils.ListColumn,
                      'members': v2_utils.ListColumn,
                      'listeners': v2_utils.ListColumn,
                      'session_persistence': v2_utils.HashColumn,
                      'tags': v2_utils.FlatListColumn}

        return (rows, (utils.get_dict_properties(
            data['pool'], rows, formatters=formatters,
            mixed_case_fields=['enable-tls'])))


class DeletePool(command.Command):
    """Delete a pool"""

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)

        parser.add_argument(
            'pool',
            metavar="<pool>",
            help="Pool to delete (name or ID)."
        )
        parser.add_argument(
            '--wait',
            action='store_true',
            help='Wait for action to complete.',
        )

        return parser

    def take_action(self, parsed_args):
        attrs = v2_utils.get_pool_attrs(self.app.client_manager, parsed_args)
        pool_id = attrs.pop('pool_id')
        self.app.client_manager.load_balancer.pool_delete(
            pool_id=pool_id)

        if parsed_args.wait:
            v2_utils.wait_for_delete(
                status_f=self.app.client_manager.load_balancer.pool_show,
                res_id=pool_id
            )


class ListPool(lister.Lister):
    """List pools"""

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)

        parser.add_argument(
            '--loadbalancer',
            metavar='<loadbalancer>',
            help="Filter by load balancer (name or ID).",
        )

        _tag.add_tag_filtering_option_to_parser(parser, 'pool')

        return parser

    def take_action(self, parsed_args):
        columns = const.POOL_COLUMNS
        attrs = v2_utils.get_pool_attrs(self.app.client_manager, parsed_args)
        data = self.app.client_manager.load_balancer.pool_list(**attrs)
        formatters = {'loadbalancers': v2_utils.ListColumn,
                      'members': v2_utils.ListColumn,
                      'listeners': v2_utils.ListColumn}

        return (columns,
                (utils.get_dict_properties(
                    s, columns, formatters=formatters) for s in data['pools']))


class ShowPool(command.ShowOne):
    """Show the details of a single pool"""

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)

        parser.add_argument(
            'pool',
            metavar='<pool>',
            help='Name or UUID of the pool.'
        )

        return parser

    def take_action(self, parsed_args):
        rows = const.POOL_ROWS
        data = None
        if uuidutils.is_uuid_like(parsed_args.pool):
            try:
                data = self.app.client_manager.load_balancer.pool_show(
                    pool_id=parsed_args.pool)
            except exceptions.NotFound:
                pass
        if data is None:
            attrs = v2_utils.get_pool_attrs(self.app.client_manager,
                                            parsed_args)
            pool_id = attrs.pop('pool_id')

            data = self.app.client_manager.load_balancer.pool_show(
                pool_id=pool_id,
            )
        formatters = {'loadbalancers': v2_utils.ListColumn,
                      'members': v2_utils.ListColumn,
                      'listeners': v2_utils.ListColumn,
                      'session_persistence': v2_utils.HashColumn,
                      'tags': v2_utils.FlatListColumn}

        return (rows, (utils.get_dict_properties(
            data, rows, formatters=formatters,
            mixed_case_fields=['enable-tls'])))


class SetPool(command.Command):
    """Update a pool"""

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)

        parser.add_argument(
            'pool',
            metavar="<pool>",
            help="Pool to update (name or ID)."
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            help="Set the name of the pool."
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help="Set the description of the pool."
        )
        parser.add_argument(
            '--session-persistence',
            metavar='<session_persistence>',
            help="Set the session persistence for the listener (key=value)."
        )
        parser.add_argument(
            '--lb-algorithm',
            metavar='{' + ','.join(ALGORITHM_CHOICES) + '}',
            choices=ALGORITHM_CHOICES,
            type=lambda s: s.upper(),  # case insensitive
            help="Set the load balancing algorithm to use."
        )
        admin_group = parser.add_mutually_exclusive_group()
        admin_group.add_argument(
            '--enable',
            action='store_true',
            default=None,
            help="Enable pool."
        )
        admin_group.add_argument(
            '--disable',
            action='store_true',
            default=None,
            help="Disable pool."
        )
        parser.add_argument(
            '--tls-container-ref',
            metavar='<container-ref>',
            help="The URI to the key manager service secrets container "
                 "containing the certificate and key for TERMINATED_TLS "
                 "pools to re-encrpt the traffic from TERMINATED_TLS "
                 "listener to backend servers."
        )
        parser.add_argument(
            '--ca-tls-container-ref',
            metavar='<ca_tls_container_ref>',
            help="The URI to the key manager service secrets container "
                 "containing the CA certificate for TERMINATED_TLS listeners "
                 "to check the backend servers certificates in ssl traffic."
        )
        parser.add_argument(
            '--crl-container-ref',
            metavar='<crl_container_ref>',
            help="The URI to the key manager service secrets container "
                 "containting the CA revocation list file for TERMINATED_TLS "
                 "listeners to valid the backend servers certificates in ssl "
                 "traffic."
        )
        tls_enable = parser.add_mutually_exclusive_group()
        tls_enable.add_argument(
            '--enable-tls',
            action='store_true',
            default=None,
            help="Enable backend associated members re-encryption."
        )
        tls_enable.add_argument(
            '--disable-tls',
            action='store_true',
            default=None,
            help="disable backend associated members re-encryption."
        )
        parser.add_argument(
            '--wait',
            action='store_true',
            help='Wait for action to complete.',
        )
        parser.add_argument(
            '--tls-ciphers',
            metavar='<tls_ciphers>',
            help="Set the TLS ciphers to be used by the pool "
                 "in OpenSSL cipher string format."
        )
        parser.add_argument(
            '--tls-version',
            dest='tls_versions',
            metavar='<tls_versions>',
            nargs='?',
            action='append',
            help="Set the TLS protocol version to be used "
                 "by the pool (can be set multiple times)."
        )
        parser.add_argument(
            '--alpn-protocol',
            dest='alpn_protocols',
            metavar='<alpn_protocols>',
            nargs='?',
            action='append',
            help="Set the ALPN protocol to be used "
                 "by the pool (can be set multiple times)."
        )

        _tag.add_tag_option_to_parser_for_set(parser, 'pool')

        return parser

    def take_action(self, parsed_args):
        attrs = v2_utils.get_pool_attrs(self.app.client_manager, parsed_args)
        pool_id = attrs.pop('pool_id')

        v2_utils.set_tags_for_set(
            self.app.client_manager.load_balancer.pool_show,
            pool_id, attrs, clear_tags=parsed_args.no_tag)

        body = {'pool': attrs}

        self.app.client_manager.load_balancer.pool_set(
            pool_id, json=body)

        if parsed_args.wait:
            v2_utils.wait_for_active(
                status_f=self.app.client_manager.load_balancer.pool_show,
                res_id=pool_id
            )


class UnsetPool(command.Command):
    """Clear pool settings"""

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)

        parser.add_argument(
            'pool',
            metavar="<pool>",
            help="Pool to modify (name or ID)."
        )
        parser.add_argument(
            '--name',
            action='store_true',
            help="Clear the pool name."
        )
        parser.add_argument(
            '--description',
            action='store_true',
            help="Clear the description of this pool."
        )
        parser.add_argument(
            '--ca-tls-container-ref',
            action='store_true',
            help="Clear the certificate authority certificate reference on "
                 "this pool."
        )
        parser.add_argument(
            '--crl-container-ref',
            action='store_true',
            help="Clear the certificate revocation list reference on "
                 "this pool."
        )
        parser.add_argument(
            '--session-persistence',
            action='store_true',
            help="Disables session persistence on the pool."
        )
        parser.add_argument(
            '--tls-container-ref',
            action='store_true',
            help="Clear the certificate reference for this pool."
        )
        parser.add_argument(
            '--tls-versions',
            action='store_true',
            help='Clear all TLS versions from the pool.',
        )
        parser.add_argument(
            '--tls-ciphers',
            action='store_true',
            help='Clear all TLS ciphers from the pool.',
        )
        parser.add_argument(
            '--wait',
            action='store_true',
            help='Wait for action to complete.',
        )
        parser.add_argument(
            '--alpn-protocols',
            action='store_true',
            help="Clear all ALPN protocols from the pool."
        )

        _tag.add_tag_option_to_parser_for_unset(parser, 'pool')

        return parser

    def take_action(self, parsed_args):
        unset_args = v2_utils.get_unsets(parsed_args)
        if not unset_args and not parsed_args.all_tag:
            return

        pool_id = v2_utils.get_resource_id(
            self.app.client_manager.load_balancer.pool_list,
            'pools', parsed_args.pool)

        v2_utils.set_tags_for_unset(
            self.app.client_manager.load_balancer.pool_show,
            pool_id, unset_args, clear_tags=parsed_args.all_tag)

        body = {'pool': unset_args}

        self.app.client_manager.load_balancer.pool_set(
            pool_id, json=body)

        if parsed_args.wait:
            v2_utils.wait_for_active(
                status_f=self.app.client_manager.load_balancer.pool_show,
                res_id=pool_id
            )



#   Copyright 2017 GoDaddy
#   Copyright 2019 Red Hat, Inc. All rights reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Health Monitor action implementation"""


from cliff import lister
from osc_lib.command import command
from osc_lib import exceptions
from osc_lib import utils
from osc_lib.utils import tags as _tag
from oslo_utils import uuidutils

from octaviaclient.osc.v2 import constants as const
from octaviaclient.osc.v2 import utils as v2_utils

HTTP_METHODS = ['GET', 'POST', 'DELETE', 'PUT', 'HEAD', 'OPTIONS', 'PATCH',
                'CONNECT', 'TRACE']
HTTP_VERSIONS = [1.0, 1.1]
TYPE_CHOICES = ['PING', 'HTTP', 'TCP', 'HTTPS', 'TLS-HELLO',
                'UDP-CONNECT', 'SCTP']


class CreateHealthMonitor(command.ShowOne):
    """Create a health monitor"""

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)

        parser.add_argument(
            'pool',
            metavar='<pool>',
            help="Set the pool for the health monitor (name or ID)."
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            help="Set the health monitor name."
        )
        parser.add_argument(
            '--delay',
            metavar='<delay>',
            required=True,
            help="Set the time in seconds, between sending probes to members."
        )
        parser.add_argument(
            '--domain-name',
            metavar='<domain_name>',
            help=("Set the domain name, which be injected into the HTTP Host "
                  "Header to the backend server for HTTP health check.")
        )
        parser.add_argument(
            '--expected-codes',
            metavar='<codes>',
            help="Set the list of HTTP status codes expected in response from "
                 "the member to declare it healthy."
        )
        parser.add_argument(
            '--http-method',
            metavar='{' + ','.join(HTTP_METHODS) + '}',
            choices=HTTP_METHODS,
            type=lambda s: s.upper(),  # case insensitive
            help="Set the HTTP method that the health monitor uses for "
                 "requests."
        )
        parser.add_argument(
            '--http-version',
            metavar='<http_version>',
            choices=HTTP_VERSIONS,
            type=float,
            help="Set the HTTP version."
        )
        parser.add_argument(
            '--timeout',
            metavar='<timeout>',
            required=True,
            help="Set the maximum time, in seconds, that a monitor waits to "
                 "connect before it times out. This value must be less than "
                 "the delay value."
        )
        parser.add_argument(
            '--max-retries',
            metavar='<max_retries>',
            type=int,
            choices=range(1, 10),
            required=True,
            help="The number of successful checks before changing the "
                 "operating status of the member to ONLINE."
        )
        parser.add_argument(
            '--url-path',
            metavar='<url_path>',
            help="Set the HTTP URL path of the request sent by the monitor to "
                 "test the health of a backend member."
        )
        parser.add_argument(
            '--type',
            metavar='{' + ','.join(TYPE_CHOICES) + '}',
            required=True,
            choices=TYPE_CHOICES,
            type=lambda s: s.upper(),  # case insensitive
            help="Set the health monitor type."
        )
        parser.add_argument(
            '--max-retries-down',
            metavar='<max_retries_down>',
            type=int,
            choices=range(1, 10),
            help="Set the number of allowed check failures before changing "
                 "the operating status of the member to ERROR."
        )
        admin_group = parser.add_mutually_exclusive_group()
        admin_group.add_argument(
            '--enable',
            action='store_true',
            default=True,
            help="Enable health monitor (default)."
        )
        admin_group.add_argument(
            '--disable',
            action='store_true',
            default=None,
            help="Disable health monitor."
        )
        parser.add_argument(
            '--wait',
            action='store_true',
            help='Wait for action to complete.',
        )

        _tag.add_tag_option_to_parser_for_create(
            parser, 'health monitor')

        return parser

    def take_action(self, parsed_args):
        rows = const.MONITOR_ROWS
        attrs = v2_utils.get_health_monitor_attrs(self.app.client_manager,
                                                  parsed_args)
        body = {"healthmonitor": attrs}
        data = self.app.client_manager.load_balancer.health_monitor_create(
            json=body)

        if parsed_args.wait:
            pool = self.app.client_manager.load_balancer.pool_show(
                data['healthmonitor']['pools'][0]['id'])
            v2_utils.wait_for_active(
                status_f=(self.app.client_manager.load_balancer.
                          load_balancer_show),
                res_id=pool['loadbalancers'][0]['id']
            )
            data = {
                'healthmonitor': (
                    self.app.client_manager.load_balancer.health_monitor_show(
                        data['healthmonitor']['id']))
            }

        formatters = {'pools': v2_utils.ListColumn,
                      'tags': v2_utils.FlatListColumn}

        return (rows,
                (utils.get_dict_properties(data['healthmonitor'],
                                           rows,
                                           formatters=formatters)))


class DeleteHealthMonitor(command.Command):
    """Delete a health monitor"""

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)

        parser.add_argument(
            'health_monitor',
            metavar='<health_monitor>',
            help="Health monitor to delete (name or ID)."
        )
        parser.add_argument(
            '--wait',
            action='store_true',
            help='Wait for action to complete.',
        )

        return parser

    def take_action(self, parsed_args):
        attrs = v2_utils.get_health_monitor_attrs(self.app.client_manager,
                                                  parsed_args)

        health_monitor_id = attrs.pop('health_monitor_id')

        self.app.client_manager.load_balancer.health_monitor_delete(
            health_monitor_id=health_monitor_id)

        if parsed_args.wait:
            v2_utils.wait_for_delete(
                status_f=(self.app.client_manager.load_balancer.
                          health_monitor_show),
                res_id=health_monitor_id
            )


class ListHealthMonitor(lister.Lister):
    """List health monitors"""

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)

        _tag.add_tag_filtering_option_to_parser(parser, 'health monitor')

        return parser

    def take_action(self, parsed_args):
        columns = const.MONITOR_COLUMNS
        attrs = v2_utils.get_health_monitor_attrs(self.app.client_manager,
                                                  parsed_args)
        data = self.app.client_manager.load_balancer.health_monitor_list(
            **attrs)

        formatters = {'pools': v2_utils.ListColumn}
        return (columns,
                (utils.get_dict_properties(s, columns, formatters=formatters)
                 for s in data['healthmonitors']))


class ShowHealthMonitor(command.ShowOne):
    """Show the details of a single health monitor"""

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)

        parser.add_argument(
            'health_monitor',
            metavar='<health_monitor>',
            help='Name or UUID of the health monitor.'
        )

        return parser

    def take_action(self, parsed_args):
        rows = const.MONITOR_ROWS
        data = None
        if uuidutils.is_uuid_like(parsed_args.health_monitor):
            try:
                data = (
                    self.app.client_manager.load_balancer.health_monitor_show(
                        health_monitor_id=parsed_args.health_monitor))
            except exceptions.NotFound:
                pass
        if data is None:
            attrs = v2_utils.get_health_monitor_attrs(self.app.client_manager,
                                                      parsed_args)

            health_monitor_id = attrs.pop('health_monitor_id')

            data = self.app.client_manager.load_balancer.health_monitor_show(
                health_monitor_id=health_monitor_id,
            )
        formatters = {'pools': v2_utils.ListColumn,
                      'tags': v2_utils.FlatListColumn}

        return (rows,
                (utils.get_dict_properties(data, rows, formatters=formatters)))


class SetHealthMonitor(command.Command):
    """Update a health monitor"""

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)

        parser.add_argument(
            'health_monitor',
            metavar='<health_monitor>',
            help="Health monitor to update (name or ID)."
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            help="Set health monitor name."
        )
        parser.add_argument(
            '--delay',
            metavar='<delay>',
            help="Set the time in seconds, between sending probes to members."
        )
        parser.add_argument(
            '--domain-name',
            metavar='<domain_name>',
            help=("Set the domain name, which be injected into the HTTP Host "
                  "Header to the backend server for HTTP health check.")
        )
        parser.add_argument(
            '--expected-codes',
            metavar='<codes>',
            help="Set the list of HTTP status codes expected in response from "
                 "the member to declare it healthy."
        )
        parser.add_argument(
            '--http-method',
            metavar='{' + ','.join(HTTP_METHODS) + '}',
            choices=HTTP_METHODS,
            type=lambda s: s.upper(),  # case insensitive
            help="Set the HTTP method that the health monitor uses for "
                 "requests."
        )
        parser.add_argument(
            '--http-version',
            metavar='<http_version>',
            choices=HTTP_VERSIONS,
            type=float,
            help="Set the HTTP version."
        )
        parser.add_argument(
            '--timeout',
            metavar='<timeout>',
            help="Set the maximum time, in seconds, that a monitor waits to "
                 "connect before it times out. This value must be less than "
                 "the delay value."
        )
        parser.add_argument(
            '--max-retries',
            metavar='<max_retries>',
            type=int,
            choices=range(1, 10),
            help="Set the number of successful checks before changing the "
                 "operating status of the member to ONLINE."
        )
        parser.add_argument(
            '--max-retries-down',
            metavar='<max_retries_down>',
            type=int,
            choices=range(1, 10),
            help="Set the number of allowed check failures before changing "
                 "the operating status of the member to ERROR."
        )
        parser.add_argument(
            '--url-path',
            metavar='<url_path>',
            help="Set the HTTP URL path of the request sent by the monitor to "
                 "test the health of a backend member."
        )
        admin_group = parser.add_mutually_exclusive_group()
        admin_group.add_argument(
            '--enable',
            action='store_true',
            default=None,
            help="Enable health monitor."
        )
        admin_group.add_argument(
            '--disable',
            action='store_true',
            default=None,
            help="Disable health monitor."
        )
        parser.add_argument(
            '--wait',
            action='store_true',
            help='Wait for action to complete.',
        )

        _tag.add_tag_option_to_parser_for_set(parser, 'health monitor')

        return parser

    def take_action(self, parsed_args):
        attrs = v2_utils.get_health_monitor_attrs(self.app.client_manager,
                                                  parsed_args)

        hm_id = attrs.pop('health_monitor_id')

        v2_utils.set_tags_for_set(
            self.app.client_manager.load_balancer.health_monitor_show,
            hm_id, attrs, clear_tags=parsed_args.no_tag)

        body = {'healthmonitor': attrs}

        self.app.client_manager.load_balancer.health_monitor_set(
            hm_id, json=body)

        if parsed_args.wait:
            v2_utils.wait_for_active(
                status_f=(self.app.client_manager.load_balancer.
                          health_monitor_show),
                res_id=hm_id
            )


class UnsetHealthMonitor(command.Command):
    """Clear health monitor settings"""

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)

        parser.add_argument(
            'health_monitor',
            metavar='<health_monitor>',
            help="Health monitor to update (name or ID)."
        )
        parser.add_argument(
            '--domain-name',
            action='store_true',
            help="Clear the health monitor domain name."
        )
        parser.add_argument(
            '--expected-codes',
            action='store_true',
            help="Reset the health monitor expected codes to the API default."
        )
        parser.add_argument(
            '--http-method',
            action='store_true',
            help="Reset the health monitor HTTP method to the API default."
        )
        parser.add_argument(
            '--http-version',
            action='store_true',
            help="Reset the health monitor HTTP version to the API default."
        )
        parser.add_argument(
            '--max-retries-down',
            action='store_true',
            help="Reset the health monitor max retries down to the API "
                 "default."
        )
        parser.add_argument(
            '--name',
            action='store_true',
            help="Clear the health monitor name."
        )
        parser.add_argument(
            '--url-path',
            action='store_true',
            help="Clear the health monitor URL path."
        )
        parser.add_argument(
            '--wait',
            action='store_true',
            help='Wait for action to complete.',
        )

        _tag.add_tag_option_to_parser_for_unset(parser, 'health monitor')

        return parser

    def take_action(self, parsed_args):
        unset_args = v2_utils.get_unsets(parsed_args)
        if not unset_args and not parsed_args.all_tag:
            return

        hm_id = v2_utils.get_resource_id(
            self.app.client_manager.load_balancer.health_monitor_list,
            'healthmonitors', parsed_args.health_monitor)

        v2_utils.set_tags_for_unset(
            self.app.client_manager.load_balancer.health_monitor_show,
            hm_id, unset_args, clear_tags=parsed_args.all_tag)

        body = {'healthmonitor': unset_args}

        self.app.client_manager.load_balancer.health_monitor_set(
            hm_id, json=body)

        if parsed_args.wait:
            v2_utils.wait_for_active(
                status_f=(self.app.client_manager.load_balancer.
                          health_monitor_show),
                res_id=hm_id
            )



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
"""OpenStackClient plugin for Load Balancer service."""
import logging

from osc_lib import utils

from octaviaclient.api.v2 import octavia

LOG = logging.getLogger(__name__)

DEFAULT_LOADBALANCER_API_VERSION = '2.0'
API_VERSION_OPTION = 'os_loadbalancer_api_version'
API_NAME = 'load_balancer'
LOAD_BALANCER_API_TYPE = 'loadbalancer'
LOAD_BALANCER_API_VERSIONS = {
    '2.0': 'octaviaclient.api.v2.octavia.OctaviaAPI',
}


def make_client(instance):
    """Returns a load balancer service client"""
    endpoint = instance.get_endpoint_for_service_type(
        'load-balancer',
        region_name=instance.region_name,
        interface=instance.interface,
    )
    client = octavia.OctaviaAPI(
        session=instance.session,
        service_type='load-balancer',
        endpoint=endpoint,
    )
    return client


def build_option_parser(parser):
    """Hook to add global options

    Called from openstackclient.shell.OpenStackShell.__init__()
    after the builtin parser has been initialized. This is
    where a plugin can add global options such as an API version.

    :param argparse.ArgumentParser parser: The parser object that
        has been initialized by OpenStackShell.
    """
    parser.add_argument(
        '--os-loadbalancer-api-version',
        metavar='<loadbalancer-api-version>',
        default=utils.env(
            'OS_LOADBALANCER_API_VERSION',
            default=DEFAULT_LOADBALANCER_API_VERSION),
        help='OSC Plugin API version, default=' +
             DEFAULT_LOADBALANCER_API_VERSION +
             ' (Env: OS_LOADBALANCER_API_VERSION)')
    return parser



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from osc_lib import exceptions
from osc_lib.tests import utils

from octaviaclient.osc.v2 import constants
from octaviaclient.osc.v2 import validate


class TestValidations(utils.TestCommand):
    def setUp(self):
        super().setUp()

    def test_check_l7policy_attrs(self):
        attrs_dict = {
            "action": "redirect_to_pool".upper(),
            "redirect_pool_id": "id"}
        try:
            validate.check_l7policy_attrs(attrs_dict)
        except exceptions.CommandError as e:
            self.fail("%s raised unexpectedly" % e)
        attrs_dict.pop("redirect_pool_id")
        self.assertRaises(
            exceptions.CommandError,
            validate.check_l7policy_attrs, attrs_dict)

        attrs_dict = {
            "action": "redirect_to_url".upper(),
            "redirect_url": "url"}
        try:
            validate.check_l7policy_attrs(attrs_dict)
        except exceptions.CommandError as e:
            self.fail("%s raised unexpectedly" % e)
        attrs_dict.pop("redirect_url")
        self.assertRaises(
            exceptions.CommandError,
            validate.check_l7policy_attrs, attrs_dict)

        attrs_dict = {
            "action": "redirect_prefix".upper(),
            "redirect_prefix": "prefix",
        }
        try:
            validate.check_l7policy_attrs(attrs_dict)
        except exceptions.CommandError as e:
            self.fail("%s raised unexpectedly" % e)
        attrs_dict.pop("redirect_prefix")
        self.assertRaises(
            exceptions.CommandError,
            validate.check_l7policy_attrs, attrs_dict)

    def test_check_l7rule_attrs(self):
        for i in ("cookie", "header"):
            attrs_dict = {
                "type": i.upper(),
                "key": "key",
            }
            try:
                validate.check_l7rule_attrs(attrs_dict)
            except exceptions.CommandError as e:
                self.fail("%s raised unexpectedly" % e)
            attrs_dict.pop("key")
            self.assertRaises(
                exceptions.CommandError,
                validate.check_l7rule_attrs, attrs_dict)

    def test_validate_TCP_UDP_SCTP_port_range(self):
        # Positive tests, should not raise
        validate._validate_TCP_UDP_SCTP_port_range(constants.MIN_PORT_NUMBER,
                                                   "fake-parameter")
        validate._validate_TCP_UDP_SCTP_port_range(constants.MAX_PORT_NUMBER,
                                                   "fake-parameter")

        # Negative tests, should raise
        self.assertRaises(exceptions.InvalidValue,
                          validate._validate_TCP_UDP_SCTP_port_range,
                          constants.MIN_PORT_NUMBER - 1, "fake-parameter")
        self.assertRaises(exceptions.InvalidValue,
                          validate._validate_TCP_UDP_SCTP_port_range,
                          constants.MAX_PORT_NUMBER + 1, "fake-parameter")

    def test_check_listener_attrs(self):
        # Positive tests, should not raise
        attrs_dict = {'protocol_port': constants.MIN_PORT_NUMBER}
        validate.check_listener_attrs(attrs_dict)
        attrs_dict = {'protocol_port': constants.MAX_PORT_NUMBER}
        validate.check_listener_attrs(attrs_dict)

        # Negative tests, should raise
        attrs_dict = {'protocol_port': constants.MIN_PORT_NUMBER - 1}
        self.assertRaises(exceptions.InvalidValue,
                          validate.check_listener_attrs, attrs_dict)
        attrs_dict = {'protocol_port': constants.MAX_PORT_NUMBER + 1}
        self.assertRaises(exceptions.InvalidValue,
                          validate.check_listener_attrs, attrs_dict)

        for key in ('hsts_preload', 'hsts_include_subdomains'):
            attrs_dict = {key: True}
            self.assertRaises(exceptions.InvalidValue,
                              validate.check_listener_attrs, attrs_dict)

    def test_check_member_attrs(self):
        # Positive tests, should not raise
        attrs_dict = {'protocol_port': constants.MIN_PORT_NUMBER,
                      'member_port': constants.MIN_PORT_NUMBER,
                      'weight': constants.MIN_WEIGHT}
        validate.check_member_attrs(attrs_dict)
        attrs_dict = {'protocol_port': constants.MAX_PORT_NUMBER,
                      'member_port': constants.MAX_PORT_NUMBER,
                      'weight': constants.MAX_WEIGHT}
        validate.check_member_attrs(attrs_dict)

        # Negative tests, should raise - protocol-port
        attrs_dict = {'protocol_port': constants.MIN_PORT_NUMBER - 1,
                      'member_port': constants.MIN_PORT_NUMBER,
                      'weight': constants.MIN_WEIGHT}
        self.assertRaises(exceptions.InvalidValue, validate.check_member_attrs,
                          attrs_dict)
        attrs_dict = {'protocol_port': constants.MAX_PORT_NUMBER + 1,
                      'member_port': constants.MIN_PORT_NUMBER,
                      'weight': constants.MIN_WEIGHT}
        self.assertRaises(exceptions.InvalidValue, validate.check_member_attrs,
                          attrs_dict)

        # Negative tests, should raise - member-port
        attrs_dict = {'protocol_port': constants.MIN_PORT_NUMBER,
                      'member_port': constants.MIN_PORT_NUMBER - 1,
                      'weight': constants.MIN_WEIGHT}
        self.assertRaises(exceptions.InvalidValue, validate.check_member_attrs,
                          attrs_dict)
        attrs_dict = {'protocol_port': constants.MIN_PORT_NUMBER,
                      'member_port': constants.MAX_PORT_NUMBER + 1,
                      'weight': constants.MIN_WEIGHT}
        self.assertRaises(exceptions.InvalidValue, validate.check_member_attrs,
                          attrs_dict)

        # Negative tests, should raise - weight
        attrs_dict = {'protocol_port': constants.MIN_PORT_NUMBER,
                      'member_port': constants.MIN_PORT_NUMBER,
                      'weight': constants.MIN_WEIGHT - 1}
        self.assertRaises(exceptions.InvalidValue, validate.check_member_attrs,
                          attrs_dict)
        attrs_dict = {'protocol_port': constants.MIN_PORT_NUMBER,
                      'member_port': constants.MIN_PORT_NUMBER,
                      'weight': constants.MAX_WEIGHT + 1}
        self.assertRaises(exceptions.InvalidValue, validate.check_member_attrs,
                          attrs_dict)



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import argparse
import copy
import itertools
from unittest import mock

from osc_lib import exceptions
from oslo_utils import uuidutils

from octaviaclient.osc.v2 import constants
from octaviaclient.osc.v2 import load_balancer
from octaviaclient.tests.unit.osc.v2 import constants as attr_consts
from octaviaclient.tests.unit.osc.v2 import fakes


class TestLoadBalancer(fakes.TestOctaviaClient):

    def setUp(self):
        super().setUp()

        self._lb = fakes.createFakeResource('loadbalancer')
        self.lb_info = copy.deepcopy(attr_consts.LOADBALANCER_ATTRS)
        self.columns = copy.deepcopy(constants.LOAD_BALANCER_COLUMNS)

        self.api_mock = mock.Mock()
        self.api_mock.load_balancer_list.return_value = copy.deepcopy(
            {'loadbalancers': [attr_consts.LOADBALANCER_ATTRS]})

        lb_client = self.app.client_manager
        lb_client.load_balancer = self.api_mock
        lb_client.neutronclient = mock.MagicMock()


class TestLoadBalancerList(TestLoadBalancer):

    def setUp(self):
        super().setUp()
        self.datalist = (tuple(
            attr_consts.LOADBALANCER_ATTRS[k] for k in self.columns),)
        self.cmd = load_balancer.ListLoadBalancer(self.app, None)

    def test_load_balancer_list_no_options(self):
        arglist = []
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)
        self.api_mock.load_balancer_list.assert_called_with()

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, tuple(data))

    def test_load_balancer_list_with_name(self):
        arglist = ['--name', 'rainbarrel']
        verifylist = [('name', 'rainbarrel')]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)
        self.api_mock.load_balancer_list.assert_called_with(name='rainbarrel')

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, tuple(data))

    @mock.patch('octaviaclient.osc.v2.utils.get_loadbalancer_attrs')
    def test_load_balancer_list_with_network(self, mock_client):
        mock_client.return_value = {
            'vip_network_id': self._lb.vip_network_id,
        }
        arglist = [
            '--vip-network-id', self._lb.vip_network_id,
        ]
        verify_list = [
            ('vip_network_id', self._lb.vip_network_id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verify_list)
        columns, data = self.cmd.take_action(parsed_args)
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, tuple(data))

    @mock.patch('octaviaclient.osc.v2.utils.get_loadbalancer_attrs')
    def test_load_balancer_list_with_subnet(self, mock_client):
        mock_client.return_value = {
            'vip_subnet_id': self._lb.vip_subnet_id,
        }
        arglist = [
            '--vip-subnet-id', self._lb.vip_subnet_id,
        ]
        verify_list = [
            ('vip_subnet_id', self._lb.vip_subnet_id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verify_list)
        columns, data = self.cmd.take_action(parsed_args)
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, tuple(data))

    @mock.patch('octaviaclient.osc.v2.utils.get_loadbalancer_attrs')
    def test_load_balancer_list_with_qos_policy(self, mock_client):
        mock_client.return_value = {
            'vip_qos_policy_id': self._lb.vip_qos_policy_id,
        }
        arglist = [
            '--vip-qos-policy-id', self._lb.vip_qos_policy_id,
        ]
        verify_list = [
            ('vip_qos_policy_id', self._lb.vip_qos_policy_id),

        ]

        parsed_args = self.check_parser(self.cmd, arglist, verify_list)
        columns, data = self.cmd.take_action(parsed_args)
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, tuple(data))

    @mock.patch('octaviaclient.osc.v2.utils.get_loadbalancer_attrs')
    def test_load_balancer_list_with_provisioning_status(self, mock_client):
        mock_client.return_value = {
            'provisioning_status': self._lb.provisioning_status,
        }
        arglist = [
            '--provisioning-status', 'active',
        ]
        verify_list = [
            ('provisioning_status', 'ACTIVE'),

        ]

        parsed_args = self.check_parser(self.cmd, arglist, verify_list)
        columns, data = self.cmd.take_action(parsed_args)
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, tuple(data))

    @mock.patch('octaviaclient.osc.v2.utils.get_loadbalancer_attrs')
    def test_load_balancer_list_with_operating_status(self, mock_client):
        mock_client.return_value = {
            'operating_status': self._lb.operating_status,
        }
        arglist = [
            '--operating-status', 'ONLiNE',
        ]
        verify_list = [
            ('operating_status', 'ONLINE'),

        ]

        parsed_args = self.check_parser(self.cmd, arglist, verify_list)
        columns, data = self.cmd.take_action(parsed_args)
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, tuple(data))

    @mock.patch('octaviaclient.osc.v2.utils.get_loadbalancer_attrs')
    def test_load_balancer_list_with_provider(self, mock_client):
        mock_client.return_value = {
            'provider': self._lb.provider,
        }
        arglist = [
            '--provider', 'octavia',
        ]
        verify_list = [
            ('provider', 'octavia'),

        ]

        parsed_args = self.check_parser(self.cmd, arglist, verify_list)
        columns, data = self.cmd.take_action(parsed_args)
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, tuple(data))

    @mock.patch('octaviaclient.osc.v2.utils.get_loadbalancer_attrs')
    def test_load_balancer_list_with_flavor(self, mock_client):
        mock_client.return_value = {
            'flavor_id': self._lb.flavor_id,
        }
        arglist = [
            '--flavor', self._lb.flavor_id,
        ]
        verify_list = [
            ('flavor', self._lb.flavor_id),

        ]

        parsed_args = self.check_parser(self.cmd, arglist, verify_list)
        columns, data = self.cmd.take_action(parsed_args)
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, tuple(data))

    @mock.patch('octaviaclient.osc.v2.utils.get_loadbalancer_attrs')
    def test_load_balancer_list_with_tags(self, mock_client):
        mock_client.return_value = {
            'tags': self._lb.tags,
        }
        arglist = [
            '--tags', ",".join(self._lb.tags),
        ]
        verify_list = [
            ('tags', self._lb.tags),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verify_list)
        columns, data = self.cmd.take_action(parsed_args)
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, tuple(data))

    @mock.patch('octaviaclient.osc.v2.utils.get_loadbalancer_attrs')
    def test_load_balancer_list_with_any_tags(self, mock_client):
        mock_client.return_value = {
            'tags': self._lb.tags,
        }
        arglist = [
            '--any-tags', ",".join(self._lb.tags),
        ]
        verify_list = [
            ('any_tags', self._lb.tags),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verify_list)
        columns, data = self.cmd.take_action(parsed_args)
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, tuple(data))

    @mock.patch('octaviaclient.osc.v2.utils.get_loadbalancer_attrs')
    def test_load_balancer_list_with_not_tags(self, mock_client):
        mock_client.return_value = {
            'tags': self._lb.tags[0],
        }
        arglist = [
            '--any-tags', ",".join(self._lb.tags),
        ]
        verify_list = [
            ('any_tags', self._lb.tags),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verify_list)
        columns, data = self.cmd.take_action(parsed_args)
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, tuple(data))

    @mock.patch('octaviaclient.osc.v2.utils.get_loadbalancer_attrs')
    def test_load_balancer_list_with_not_any_tags(self, mock_client):
        mock_client.return_value = {
            'tags': self._lb.tags[0],
        }
        arglist = [
            '--not-any-tags', ",".join(self._lb.tags),
        ]
        verify_list = [
            ('not_any_tags', self._lb.tags),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verify_list)
        columns, data = self.cmd.take_action(parsed_args)
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, tuple(data))


class TestLoadBalancerDelete(TestLoadBalancer):

    def setUp(self):
        super().setUp()
        self.cmd = load_balancer.DeleteLoadBalancer(self.app, None)

    def test_load_balancer_delete(self):
        arglist = [self._lb.id]
        verifylist = [
            ('loadbalancer', self._lb.id)
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)
        self.api_mock.load_balancer_delete.assert_called_with(
            lb_id=self._lb.id)

    @mock.patch('osc_lib.utils.wait_for_delete')
    def test_load_balancer_delete_wait(self, mock_wait):
        arglist = [self._lb.id, '--wait']
        verifylist = [
            ('loadbalancer', self._lb.id),
            ('wait', True),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)
        self.api_mock.load_balancer_delete.assert_called_with(
            lb_id=self._lb.id)
        mock_wait.assert_called_once_with(
            manager=mock.ANY,
            res_id=self.lb_info['id'],
            sleep_time=mock.ANY,
            status_field='provisioning_status')

    def test_load_balancer_delete_failure(self):
        arglist = ['unknown_lb']
        verifylist = [
            ('loadbalancer', 'unknown_lb')
        ]
        self.api_mock.load_balancer_list.return_value = {
            'loadbalancers': []}
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.assertRaises(exceptions.CommandError, self.cmd.take_action,
                          parsed_args)
        self.assertNotCalled(self.api_mock.load_balancer_delete)


class TestLoadBalancerCreate(TestLoadBalancer):

    def setUp(self):
        super().setUp()

        self.api_mock.load_balancer_create.return_value = {
            'loadbalancer': self.lb_info
        }
        lb_client = self.app.client_manager
        lb_client.load_balancer = self.api_mock

        self.cmd = load_balancer.CreateLoadBalancer(self.app, None)

    @mock.patch('octaviaclient.osc.v2.utils.get_loadbalancer_attrs')
    def test_load_balancer_create(self, mock_client):
        mock_client.return_value = self.lb_info
        arglist = ['--name', self._lb.name,
                   '--vip-network-id', self._lb.vip_network_id,
                   '--project', self._lb.project_id,
                   '--flavor', self._lb.flavor_id]
        verifylist = [
            ('name', self._lb.name),
            ('vip_network_id', self._lb.vip_network_id),
            ('project', self._lb.project_id),
            ('flavor', self._lb.flavor_id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)
        self.api_mock.load_balancer_create.assert_called_with(
            json={'loadbalancer': self.lb_info})

    @mock.patch('osc_lib.utils.wait_for_status')
    @mock.patch('octaviaclient.osc.v2.utils.get_loadbalancer_attrs')
    def test_load_balancer_create_wait(self, mock_client, mock_wait):
        mock_client.return_value = self.lb_info
        self.api_mock.load_balancer_show.return_value = self.lb_info
        arglist = ['--name', self._lb.name,
                   '--vip-network-id', self._lb.vip_network_id,
                   '--project', self._lb.project_id,
                   '--flavor', self._lb.flavor_id,
                   '--wait']
        verifylist = [
            ('name', self._lb.name),
            ('vip_network_id', self._lb.vip_network_id),
            ('project', self._lb.project_id),
            ('flavor', self._lb.flavor_id),
            ('wait', True),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)
        self.api_mock.load_balancer_create.assert_called_with(
            json={'loadbalancer': self.lb_info})
        mock_wait.assert_called_once_with(
            status_f=mock.ANY,
            res_id=self.lb_info['id'],
            sleep_time=mock.ANY,
            status_field='provisioning_status')

    @mock.patch('octaviaclient.osc.v2.utils.get_loadbalancer_attrs')
    def test_load_balancer_create_with_qos_policy(self, mock_client):
        qos_policy_id = 'qos_id'
        lb_info = copy.deepcopy(self.lb_info)
        lb_info.update({'vip_qos_policy_id': qos_policy_id})
        mock_client.return_value = lb_info

        arglist = [
            '--name', self._lb.name,
            '--vip-network-id', self._lb.vip_network_id,
            '--project', self._lb.project_id,
            '--vip-qos-policy-id', qos_policy_id,
        ]
        verifylist = [
            ('name', self._lb.name),
            ('vip_network_id', self._lb.vip_network_id),
            ('project', self._lb.project_id),
            ('vip_qos_policy_id', qos_policy_id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)
        self.api_mock.load_balancer_create.assert_called_with(
            json={'loadbalancer': lb_info})

    @mock.patch('octaviaclient.osc.v2.utils.get_loadbalancer_attrs')
    def test_load_balancer_create_with_additional_vips(self, mock_client):
        mock_client.return_value = self.lb_info

        arglist = [
            '--name', self._lb.name,
            '--vip-subnet-id', self._lb.vip_subnet_id,
            '--project', self._lb.project_id,
            '--additional-vip', 'subnet-id={},ip-address={}'.format(
                self._lb.additional_vips[0]['subnet_id'],
                self._lb.additional_vips[0]['ip_address']),
            '--additional-vip', 'subnet-id={},ip-address={}'.format(
                self._lb.additional_vips[1]['subnet_id'],
                self._lb.additional_vips[1]['ip_address'])
        ]
        verifylist = [
            ('name', self._lb.name),
            ('vip_subnet_id', self._lb.vip_subnet_id),
            ('project', self._lb.project_id),
            ('additional_vip', [
                'subnet-id={},ip-address={}'.format(
                    self._lb.additional_vips[0]['subnet_id'],
                    self._lb.additional_vips[0]['ip_address']),
                'subnet-id={},ip-address={}'.format(
                    self._lb.additional_vips[1]['subnet_id'],
                    self._lb.additional_vips[1]['ip_address'])])
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)
        self.api_mock.load_balancer_create.assert_called_with(
            json={'loadbalancer': self.lb_info})

    @mock.patch('octaviaclient.osc.v2.utils.get_loadbalancer_attrs')
    def test_load_balancer_create_with_provider(self, mock_client):
        provider = 'foobar'
        lb_info = copy.deepcopy(self.lb_info)
        lb_info.update({'provider': provider})
        mock_client.return_value = lb_info

        arglist = [
            '--name', self._lb.name,
            '--vip-network-id', self._lb.vip_network_id,
            '--project', self._lb.project_id,
            '--provider', provider,
        ]
        verifylist = [
            ('name', self._lb.name),
            ('vip_network_id', self._lb.vip_network_id),
            ('project', self._lb.project_id),
            ('provider', provider),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)
        self.api_mock.load_balancer_create.assert_called_with(
            json={'loadbalancer': lb_info})

    @mock.patch('octaviaclient.osc.v2.utils.get_loadbalancer_attrs')
    def test_load_balancer_create_with_vip_sg_ids(self, mock_client):
        vip_sg_ids = [
            uuidutils.generate_uuid(),
            uuidutils.generate_uuid()
        ]
        lb_info = copy.deepcopy(self.lb_info)
        lb_info.update({'vip_sg_ids': vip_sg_ids})
        mock_client.return_value = lb_info

        arglist = [
            '--name', self._lb.name,
            '--vip-network-id', self._lb.vip_network_id,
            '--project', self._lb.project_id,
            '--vip-sg-id', vip_sg_ids[0],
            '--vip-sg-id', vip_sg_ids[1],
        ]
        verifylist = [
            ('name', self._lb.name),
            ('vip_network_id', self._lb.vip_network_id),
            ('project', self._lb.project_id),
            ('vip_sg_id', vip_sg_ids),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)
        self.api_mock.load_balancer_create.assert_called_with(
            json={'loadbalancer': lb_info})

    @mock.patch('octaviaclient.osc.v2.utils.get_loadbalancer_attrs')
    def test_load_balancer_create_with_tags(self, mock_client):
        lb_info = copy.deepcopy(self.lb_info)
        lb_info.update({'tags': self._lb.tags})
        mock_client.return_value = lb_info

        arglist = [
            '--name', self._lb.name,
            '--vip-network-id', self._lb.vip_network_id,
            '--project', self._lb.project_id,
            '--tag', self._lb.tags[0],
            '--tag', self._lb.tags[1],
        ]
        verifylist = [
            ('name', self._lb.name),
            ('vip_network_id', self._lb.vip_network_id),
            ('project', self._lb.project_id),
            ('tags', self._lb.tags),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)
        self.api_mock.load_balancer_create.assert_called_with(
            json={'loadbalancer': lb_info})

    @mock.patch('octaviaclient.osc.v2.utils.get_loadbalancer_attrs')
    def test_load_balancer_create_missing_args(self, mock_client):
        attrs_list = self.lb_info

        # init missing keys
        args = ("vip_subnet_id", "vip_network_id", "vip_port_id")
        for a in args:
            # init missing keys
            attrs_list[a] = ''
        # verify all valid combinations of args
        for n in range(len(args) + 1):
            for comb in itertools.combinations(args, n):
                # subtract comb's keys from attrs_list
                filtered_attrs = {k: v for k, v in attrs_list.items() if (
                    k not in comb)}

                # Add the 'wait' attribute, which isn't part of an LB directly
                filtered_attrs['wait'] = False
                mock_client.return_value = filtered_attrs
                parsed_args = argparse.Namespace(**filtered_attrs)

                if not any(k in filtered_attrs for k in args) or all(
                    k in filtered_attrs for k in ("vip_network_id",
                                                  "vip_port_id")
                ):
                    self.assertRaises(
                        exceptions.CommandError,
                        self.cmd.take_action,
                        parsed_args
                    )
                else:
                    try:
                        self.cmd.take_action(parsed_args)
                    except exceptions.CommandError as e:
                        self.fail("%s raised unexpectedly" % e)


class TestLoadBalancerShow(TestLoadBalancer):

    def setUp(self):
        super().setUp()
        self.api_mock.load_balancer_show.return_value = {
            'loadbalancer': self.lb_info}
        lb_client = self.app.client_manager
        lb_client.load_balancer = self.api_mock

        self.cmd = load_balancer.ShowLoadBalancer(self.app, None)

    def test_load_balancer_show(self):
        arglist = [self._lb.id]
        verifylist = [
            ('loadbalancer', self._lb.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)
        self.api_mock.load_balancer_show.assert_called_with(lb_id=self._lb.id)


class TestLoadBalancerSet(TestLoadBalancer):

    def setUp(self):
        super().setUp()
        lb_client = self.app.client_manager
        lb_client.load_balancer = self.api_mock
        self.cmd = load_balancer.SetLoadBalancer(self.app, None)

    @mock.patch('octaviaclient.osc.v2.utils.get_loadbalancer_attrs')
    def test_load_balancer_set(self, mock_attrs):
        qos_policy_id = uuidutils.generate_uuid()
        mock_attrs.return_value = {
            'loadbalancer_id': self._lb.id,
            'name': 'new_name',
            'vip_qos_policy_id': qos_policy_id,
        }
        arglist = [self._lb.id, '--name', 'new_name',
                   '--vip-qos-policy-id', qos_policy_id]
        verifylist = [
            ('loadbalancer', self._lb.id),
            ('name', 'new_name'),
            ('vip_qos_policy_id', qos_policy_id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)
        self.api_mock.load_balancer_set.assert_called_with(
            self._lb.id, json={
                'loadbalancer': {
                    'name': 'new_name',
                    'vip_qos_policy_id': qos_policy_id,
                }
            })

    @mock.patch('osc_lib.utils.wait_for_status')
    @mock.patch('octaviaclient.osc.v2.utils.get_loadbalancer_attrs')
    def test_load_balancer_set_wait(self, mock_attrs, mock_wait):
        qos_policy_id = uuidutils.generate_uuid()
        mock_attrs.return_value = {
            'loadbalancer_id': self._lb.id,
            'name': 'new_name',
            'vip_qos_policy_id': qos_policy_id,
        }
        arglist = [self._lb.id, '--name', 'new_name',
                   '--vip-qos-policy-id', qos_policy_id, '--wait']
        verifylist = [
            ('loadbalancer', self._lb.id),
            ('name', 'new_name'),
            ('vip_qos_policy_id', qos_policy_id),
            ('wait', True),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)
        self.api_mock.load_balancer_set.assert_called_with(
            self._lb.id, json={
                'loadbalancer': {
                    'name': 'new_name',
                    'vip_qos_policy_id': qos_policy_id,
                }
            })
        mock_wait.assert_called_once_with(
            status_f=mock.ANY,
            res_id=self.lb_info['id'],
            sleep_time=mock.ANY,
            status_field='provisioning_status')

    @mock.patch('octaviaclient.osc.v2.utils.get_loadbalancer_attrs')
    def test_load_balancer_set_tag(self, mock_attrs):
        self.api_mock.load_balancer_show.return_value = {
            'tags': ['foo']
        }

        mock_attrs.return_value = {
            'loadbalancer_id': self._lb.id,
            'tags': ['bar']
        }
        arglist = [self._lb.id, '--tag', 'bar']
        verifylist = [
            ('loadbalancer', self._lb.id),
            ('tags', ['bar'])
        ]

        try:
            parsed_args = self.check_parser(self.cmd, arglist, verifylist)
            self.cmd.take_action(parsed_args)
        except Exception as e:
            self.fail("%s raised unexpectedly" % e)

        self.api_mock.load_balancer_set.assert_called_once()
        kwargs = self.api_mock.load_balancer_set.mock_calls[0][2]
        tags = kwargs['json']['loadbalancer']['tags']
        self.assertEqual(2, len(tags))
        self.assertIn('foo', tags)
        self.assertIn('bar', tags)

    @mock.patch('octaviaclient.osc.v2.utils.get_loadbalancer_attrs')
    def test_load_balancer_set_tag_no_tag(self, mock_attrs):
        self.api_mock.load_balancer_show.return_value = {
            'tags': ['foo']
        }

        mock_attrs.return_value = {
            'loadbalancer_id': self._lb.id,
            'tags': ['bar']
        }
        arglist = [self._lb.id, '--tag', 'bar', '--no-tag']
        verifylist = [
            ('loadbalancer', self._lb.id),
            ('tags', ['bar'])
        ]

        try:
            parsed_args = self.check_parser(self.cmd, arglist, verifylist)
            self.cmd.take_action(parsed_args)
        except Exception as e:
            self.fail("%s raised unexpectedly" % e)

        self.api_mock.load_balancer_set.assert_called_once_with(
            self._lb.id,
            json={'loadbalancer': {'tags': ['bar']}})

    @mock.patch('octaviaclient.osc.v2.utils.get_loadbalancer_attrs')
    def test_load_balancer_remove_qos_policy(self, mock_attrs):
        mock_attrs.return_value = {
            'loadbalancer_id': self._lb.id,
            'vip_qos_policy_id': None,
        }
        arglist = [self._lb.id, '--vip-qos-policy-id', 'None']
        verifylist = [
            ('loadbalancer', self._lb.id),
            ('vip_qos_policy_id', 'None'),
        ]

        try:
            parsed_args = self.check_parser(self.cmd, arglist, verifylist)
            self.cmd.take_action(parsed_args)
        except Exception as e:
            self.fail("%s raised unexpectedly" % e)

    @mock.patch('octaviaclient.osc.v2.utils.get_loadbalancer_attrs')
    def test_load_balancer_set_vip_sg_id(self, mock_attrs):
        mock_attrs.return_value = {
            'loadbalancer_id': self._lb.id,
        }
        sg_id = uuidutils.generate_uuid()
        arglist = [self._lb.id, '--vip-sg-id', sg_id]
        verifylist = [
            ('loadbalancer', self._lb.id),
            ('vip_sg_id', [sg_id]),
        ]

        try:
            parsed_args = self.check_parser(self.cmd, arglist, verifylist)
            self.cmd.take_action(parsed_args)
        except Exception as e:
            self.fail("%s raised unexpectedly" % e)


class TestLoadBalancerStats(TestLoadBalancer):

    def setUp(self):
        super().setUp()
        lb_stats_info = {'stats': {'bytes_in': '0'}}
        self.api_mock.load_balancer_stats_show.return_value = {
            'stats': lb_stats_info['stats']}
        lb_client = self.app.client_manager
        lb_client.load_balancer = self.api_mock
        self.cmd = load_balancer.ShowLoadBalancerStats(self.app, None)

    def test_load_balancer_stats_show(self):
        arglist = [self._lb.id]
        verifylist = [
            ('loadbalancer', self._lb.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)
        self.api_mock.load_balancer_stats_show.assert_called_with(
            lb_id=self._lb.id)


class TestLoadBalancerStatus(TestLoadBalancer):
    def setUp(self):
        super().setUp()
        expected_res = {'statuses': {'operating_status': 'ONLINE',
                                     'provisioning_status': 'ACTIVE'}}
        self.api_mock.load_balancer_status_show.return_value = {
            'statuses': expected_res['statuses']
        }
        lb_client = self.app.client_manager
        lb_client.load_balancer = self.api_mock
        self.cmd = load_balancer.ShowLoadBalancerStatus(self.app, None)

    def test_load_balancer_status_show(self):
        # lbaas-loadbalancer-status test_id.
        arglist = [self._lb.id]
        verifylist = [
            ('loadbalancer', self._lb.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)
        self.api_mock.load_balancer_status_show.assert_called_with(
            lb_id=self._lb.id)


class TestLoadBalancerFailover(TestLoadBalancer):

    def setUp(self):
        super().setUp()
        self.cmd = load_balancer.FailoverLoadBalancer(self.app, None)

    def test_load_balancer_failover(self):
        arglist = [self._lb.id]
        verifylist = [
            ('loadbalancer', self._lb.id)
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)
        self.api_mock.load_balancer_failover.assert_called_with(
            lb_id=self._lb.id)

    @mock.patch('osc_lib.utils.wait_for_status')
    def test_load_balancer_failover_wait(self, mock_wait):
        arglist = [self._lb.id, '--wait']
        verifylist = [
            ('loadbalancer', self._lb.id),
            ('wait', True),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)
        self.api_mock.load_balancer_failover.assert_called_with(
            lb_id=self._lb.id)
        mock_wait.assert_called_once_with(
            status_f=mock.ANY,
            res_id=self._lb.id,
            sleep_time=mock.ANY,
            status_field='provisioning_status')


class TestLoadBalancerUnset(TestLoadBalancer):
    PARAMETERS = ('name', 'description', 'vip_qos_policy_id')

    def setUp(self):
        super().setUp()
        lb_client = self.app.client_manager
        lb_client.load_balancer = self.api_mock
        self.cmd = load_balancer.UnsetLoadBalancer(self.app, None)

    def test_load_balancer_unset_name(self):
        self._test_load_balancer_unset_param('name')

    def test_load_balancer_unset_name_wait(self):
        self._test_load_balancer_unset_param_wait('name')

    def test_load_balancer_unset_description(self):
        self._test_load_balancer_unset_param('description')

    def test_load_balancer_unset_qos(self):
        self._test_load_balancer_unset_param('vip_qos_policy_id')

    def _test_load_balancer_unset_param(self, param):
        self.api_mock.load_balancer_set.reset_mock()
        ref_body = {'loadbalancer': {param: None}}
        arg_param = param.replace('_', '-') if '_' in param else param
        arglist = [self._lb.id, '--%s' % arg_param]
        verifylist = [
            ('loadbalancer', self._lb.id),
        ]
        for ref_param in self.PARAMETERS:
            verifylist.append((ref_param, param == ref_param))
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)
        self.api_mock.load_balancer_set.assert_called_once_with(
            self._lb.id, json=ref_body)

    @mock.patch('osc_lib.utils.wait_for_status')
    def _test_load_balancer_unset_param_wait(self, param, mock_wait):
        self.api_mock.load_balancer_set.reset_mock()
        ref_body = {'loadbalancer': {param: None}}
        arg_param = param.replace('_', '-') if '_' in param else param
        arglist = [self._lb.id, '--%s' % arg_param, '--wait']
        verifylist = [
            ('loadbalancer', self._lb.id),
            ('wait', True),
        ]
        for ref_param in self.PARAMETERS:
            verifylist.append((ref_param, param == ref_param))
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)
        self.api_mock.load_balancer_set.assert_called_once_with(
            self._lb.id, json=ref_body)
        mock_wait.assert_called_once_with(
            status_f=mock.ANY,
            res_id=self.lb_info['id'],
            sleep_time=mock.ANY,
            status_field='provisioning_status')

    def test_load_balancer_unset_all(self):
        self.api_mock.load_balancer_set.reset_mock()
        ref_body = {'loadbalancer': {x: None for x in self.PARAMETERS}}
        arglist = [self._lb.id]
        for ref_param in self.PARAMETERS:
            arg_param = (ref_param.replace('_', '-') if '_' in ref_param else
                         ref_param)
            arglist.append('--%s' % arg_param)
        verifylist = list(zip(self.PARAMETERS, [True] * len(self.PARAMETERS)))
        verifylist = [('loadbalancer', self._lb.id)] + verifylist
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)
        self.api_mock.load_balancer_set.assert_called_once_with(
            self._lb.id, json=ref_body)

    def test_load_balancer_unset_none(self):
        self.api_mock.load_balancer_set.reset_mock()
        arglist = [self._lb.id]
        verifylist = list(zip(self.PARAMETERS, [False] * len(self.PARAMETERS)))
        verifylist = [('loadbalancer', self._lb.id)] + verifylist
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)
        self.api_mock.load_balancer_set.assert_not_called()

    def test_load_balancer_unset_tag(self):
        self.api_mock.load_balancer_show.return_value = {
            'tags': ['foo', 'bar']
        }

        arglist = [self._lb.id, '--tag', 'foo']
        verifylist = [
            ('loadbalancer', self._lb.id),
            ('tags', ['foo'])
        ]

        try:
            parsed_args = self.check_parser(self.cmd, arglist, verifylist)
            self.cmd.take_action(parsed_args)
        except Exception as e:
            self.fail("%s raised unexpectedly" % e)

        self.api_mock.load_balancer_set.assert_called_once_with(
            self._lb.id,
            json={'loadbalancer': {'tags': ['bar']}})

    def test_load_balancer_unset_all_tag(self):
        self.api_mock.load_balancer_show.return_value = {
            'tags': ['foo', 'bar']
        }

        arglist = [self._lb.id, '--all-tag']
        verifylist = [
            ('loadbalancer', self._lb.id),
            ('all_tag', True)
        ]

        try:
            parsed_args = self.check_parser(self.cmd, arglist, verifylist)
            self.cmd.take_action(parsed_args)
        except Exception as e:
            self.fail("%s raised unexpectedly" % e)

        self.api_mock.load_balancer_set.assert_called_once_with(
            self._lb.id,
            json={'loadbalancer': {'tags': []}})

    @mock.patch('octaviaclient.osc.v2.utils.get_loadbalancer_attrs')
    def test_load_balancer_unset_vip_sg_id(self, mock_attrs):
        sg_id = uuidutils.generate_uuid()
        mock_attrs.return_value = {
            'loadbalancer_id': self._lb.id,
            'vip_sg_ids': [sg_id]
        }
        arglist = [self._lb.id, '--vip-sg-id']
        verifylist = [
            ('loadbalancer', self._lb.id),
            ('vip_sg_ids', True),
        ]

        try:
            parsed_args = self.check_parser(self.cmd, arglist, verifylist)
            self.cmd.take_action(parsed_args)
        except Exception as e:
            self.fail("%s raised unexpectedly" % e)

        self.api_mock.load_balancer_set.assert_called_once_with(
            self._lb.id,
            json={'loadbalancer': {'vip_sg_ids': None}})



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Amphora action implementation"""


from cliff import lister
from osc_lib.command import command
from osc_lib import utils

from octaviaclient.osc.v2 import constants as const
from octaviaclient.osc.v2 import utils as v2_utils


class ListAmphora(lister.Lister):
    """List amphorae"""

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)

        parser.add_argument(
            '--loadbalancer',
            metavar='<loadbalancer>',
            dest='loadbalancer',
            help="Filter by load balancer (name or ID).",
        )
        parser.add_argument(
            '--compute-id',
            metavar='<compute-id>',
            help="Filter by compute ID.",
        )

        role_choices = {'MASTER', 'BACKUP', 'STANDALONE'}
        parser.add_argument(
            '--role',
            metavar='{' + ','.join(sorted(role_choices)) + '}',
            choices=role_choices,
            type=lambda s: s.upper(),  # case insensitive
            help="Filter by role."
        )

        status_choices = {
            'ALLOCATED', 'BOOTING', 'DELETED', 'ERROR',
            'PENDING_CREATE', 'PENDING_DELETE', 'READY', 'FAILOVER_STOPPED',
        }
        parser.add_argument(
            '--status', '--provisioning-status',
            dest='status',
            metavar='{' + ','.join(sorted(status_choices)) + '}',
            choices=status_choices,
            type=lambda s: s.upper(),  # case insensitive
            help="Filter by amphora provisioning status."
        )

        parser.add_argument(
            '--image-id',
            metavar='<image-id>',
            help="Filter by image ID.",
        )

        parser.add_argument(
            '--long',
            action='store_true',
            help='Show additional fields.',
        )

        return parser

    def take_action(self, parsed_args):
        columns = const.AMPHORA_COLUMNS
        if parsed_args.long:
            columns = const.AMPHORA_COLUMNS_LONG

        attrs = v2_utils.get_amphora_attrs(self.app.client_manager,
                                           parsed_args)

        data = self.app.client_manager.load_balancer.amphora_list(**attrs)

        formatters = {
            'amphorae': v2_utils.ListColumn,
        }

        return (
            columns,
            (utils.get_dict_properties(
                amp,
                columns,
                formatters=formatters) for amp in data['amphorae']),
        )


class ShowAmphora(command.ShowOne):
    """Show the details of a single amphora"""

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)

        parser.add_argument(
            'amphora_id',
            metavar='<amphora-id>',
            help='UUID of the amphora.',
        )

        return parser

    def take_action(self, parsed_args):

        data = self.app.client_manager.load_balancer.amphora_show(
            amphora_id=parsed_args.amphora_id,
        )

        rows = const.AMPHORA_ROWS
        formatters = {
            'loadbalancers': v2_utils.ListColumn,
            'amphorae': v2_utils.ListColumn,
        }

        return (rows, utils.get_dict_properties(data, rows,
                                                formatters=formatters))


class ConfigureAmphora(command.Command):
    """Update the amphora agent configuration"""

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)

        parser.add_argument(
            'amphora_id',
            metavar='<amphora-id>',
            help='UUID of the amphora to configure.',
        )
        parser.add_argument(
            '--wait',
            action='store_true',
            help='Wait for action to complete.',
        )

        return parser

    def take_action(self, parsed_args):
        attrs = v2_utils.get_amphora_attrs(self.app.client_manager,
                                           parsed_args)
        amp_id = attrs.pop('amphora_id')
        self.app.client_manager.load_balancer.amphora_configure(
            amphora_id=amp_id)

        if parsed_args.wait:
            amphora = self.app.client_manager.load_balancer.amphora_show(
                amp_id)
            lb_id = amphora.get('loadbalancer_id')
            # TODO(rm_work): No status change if the amp isn't linked to an LB?
            if lb_id:
                v2_utils.wait_for_active(
                    status_f=(self.app.client_manager.load_balancer.
                              load_balancer_show),
                    res_id=lb_id
                )


class FailoverAmphora(command.Command):
    """Force failover an amphora"""

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)

        parser.add_argument(
            'amphora_id',
            metavar='<amphora-id>',
            help='UUID of the amphora.',
        )
        parser.add_argument(
            '--wait',
            action='store_true',
            help='Wait for action to complete.',
        )

        return parser

    def take_action(self, parsed_args):
        attrs = v2_utils.get_amphora_attrs(self.app.client_manager,
                                           parsed_args)
        amp_id = attrs.pop('amphora_id')
        amphora = self.app.client_manager.load_balancer.amphora_show(amp_id)
        self.app.client_manager.load_balancer.amphora_failover(
            amphora_id=amp_id)

        if parsed_args.wait:
            lb_id = amphora.get('loadbalancer_id')
            if lb_id:
                v2_utils.wait_for_active(
                    status_f=(self.app.client_manager.load_balancer.
                              load_balancer_show),
                    res_id=lb_id
                )
            else:
                v2_utils.wait_for_delete(
                    status_f=(self.app.client_manager.load_balancer.
                              amphora_show),
                    res_id=amp_id
                )


class ShowAmphoraStats(command.ShowOne):
    """Shows the current statistics for an amphora."""

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)

        parser.add_argument(
            '--listener',
            metavar='<listener>',
            help='Filter by listener (name or ID).',
        )
        parser.add_argument(
            'amphora_id',
            metavar='<amphora-id>',
            help='UUID of the amphora.'
        )

        return parser

    def take_action(self, parsed_args):
        rows = const.LOAD_BALANCER_STATS_ROWS

        listener_id = None

        if parsed_args.listener is not None:
            attrs = v2_utils.get_listener_attrs(
                self.app.client_manager,
                parsed_args)
            listener_id = attrs.pop('listener_id')

        data = self.app.client_manager.load_balancer.amphora_stats_show(
            amphora_id=parsed_args.amphora_id
        )

        total_stats = {
            key: 0
            for key in rows
        }
        for stats in data['amphora_stats']:
            if listener_id is None or listener_id == stats['listener_id']:
                for key in stats:
                    if key in rows:
                        total_stats[key] += stats[key]

        return (rows, (utils.get_dict_properties(
            total_stats, rows, formatters={})))


class DeleteAmphora(command.Command):
    """Delete a amphora"""

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)

        parser.add_argument(
            'amphora_id',
            metavar='<amphora-id>',
            help='UUID of the amphora to delete.',
        )
        parser.add_argument(
            '--wait',
            action='store_true',
            help='Wait for action to complete.',
        )

        return parser

    def take_action(self, parsed_args):

        self.app.client_manager.load_balancer.amphora_delete(
            amphora_id=parsed_args.amphora_id)

        if parsed_args.wait:
            v2_utils.wait_for_delete(
                status_f=self.app.client_manager.load_balancer.amphora_show,
                res_id=parsed_args.amphora_id, status_field=const.STATUS
            )



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#
import copy
from unittest import mock

from osc_lib.tests import utils

from octaviaclient.tests import fakes
from octaviaclient.tests.unit.osc.v2 import constants


class FakeOctaviaClient(object):
    def __init__(self, **kwargs):
        self.load_balancers = mock.Mock()
        self.load_balancers.resource_class = fakes.FakeResource(None, {})
        self.auth_token = kwargs['token']
        self.management_url = kwargs['endpoint']


class TestOctaviaClient(utils.TestCommand):

    def setUp(self):
        super().setUp()
        self.app.client_manager.load_balancer = FakeOctaviaClient(
            endpoint=fakes.AUTH_URL,
            token=fakes.AUTH_TOKEN,
        )


def createFakeResource(name, attrs=None):
    """Creates a single fake resource object.

    :param name: resource_name
    :param attrs: ``dict`` of customized resource attributes
    :returns: A FakeResource object
    """
    attrs = attrs or {}

    # Set to default
    resource_info = getattr(constants, "{}_attrs".format(name).upper())
    assert resource_info is not None, "{} is not found".format(name)

    resource_info.update(attrs)
    return fakes.FakeResource(
        info=copy.deepcopy(resource_info), loaded=True,
    )



#   Copyright 2019 Red Hat, Inc. All rights reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Member action implementation"""

import functools

from cliff import lister
from osc_lib.command import command
from osc_lib import exceptions
from osc_lib import utils
from osc_lib.utils import tags as _tag
from oslo_utils import uuidutils

from octaviaclient.osc.v2 import constants as const
from octaviaclient.osc.v2 import utils as v2_utils
from octaviaclient.osc.v2 import validate


class ListMember(lister.Lister):
    """List members in a pool"""

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)

        parser.add_argument(
            'pool',
            metavar='<pool>',
            help="Pool name or ID to list the members of."
        )

        _tag.add_tag_filtering_option_to_parser(parser, 'member')

        return parser

    def take_action(self, parsed_args):
        columns = const.MEMBER_COLUMNS

        attrs = v2_utils.get_member_attrs(self.app.client_manager, parsed_args)

        data = self.app.client_manager.load_balancer.member_list(
            **attrs)

        return (columns,
                (utils.get_dict_properties(
                    s, columns,
                    formatters={},
                ) for s in data['members']))


class ShowMember(command.ShowOne):
    """Shows details of a single Member"""

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)

        parser.add_argument(
            'pool',
            metavar='<pool>',
            help='Pool name or ID to show the members of.'
        )
        parser.add_argument(
            'member',
            metavar='<member>',
            help="Name or ID of the member to show."
        )

        return parser

    def take_action(self, parsed_args):
        rows = const.MEMBER_ROWS
        data = None
        if (uuidutils.is_uuid_like(parsed_args.pool) and
                uuidutils.is_uuid_like(parsed_args.member)):
            try:
                data = self.app.client_manager.load_balancer.member_show(
                    pool_id=parsed_args.pool, member_id=parsed_args.member)
            except exceptions.NotFound:
                pass
        if data is None:
            attrs = v2_utils.get_member_attrs(self.app.client_manager,
                                              parsed_args)

            member_id = attrs.pop('member_id')
            pool_id = attrs.pop('pool_id')

            data = self.app.client_manager.load_balancer.member_show(
                pool_id=pool_id, member_id=member_id)

        # Handle older API versions that did not have the vnic_type
        if not data.get('vnic_type', False):
            data['vnic_type'] = 'normal'

        formatters = {'tags': v2_utils.FlatListColumn}

        return (rows, (utils.get_dict_properties(
            data, rows, formatters=formatters)))


class CreateMember(command.ShowOne):
    """Creating a member in a pool"""

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)

        parser.add_argument(
            'pool',
            metavar='<pool>',
            help="ID or name of the pool to create the member for."
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            help="Name of the member."
        )
        backup = parser.add_mutually_exclusive_group()
        backup.add_argument(
            '--disable-backup',
            action='store_true',
            default=None,
            help="Disable member backup (default)."
        )
        backup.add_argument(
            '--enable-backup',
            action='store_true',
            default=None,
            help="Enable member backup."
        )
        parser.add_argument(
            '--weight',
            metavar='<weight>',
            type=int,
            help="The weight of a member determines the portion of requests "
                 "or connections it services compared to the other members of "
                 "the pool."
        )
        parser.add_argument(
            '--address',
            metavar='<ip_address>',
            help="The IP address of the backend member server.",
            required=True
        )
        parser.add_argument(
            '--subnet-id',
            metavar='<subnet_id>',
            help="The subnet ID the member service is accessible from."
        )
        parser.add_argument(
            '--protocol-port',
            metavar='<protocol_port>',
            type=int,
            help="The protocol port number the backend member server is "
                 "listening on.",
            required=True
        )
        parser.add_argument(
            '--monitor-port',
            metavar='<monitor_port>',
            type=int,
            help="An alternate protocol port used for health monitoring a "
                 "backend member.",
        )
        parser.add_argument(
            '--monitor-address',
            metavar='<monitor_address>',
            help="An alternate IP address used for health monitoring a "
                 "backend member."
        )
        admin_group = parser.add_mutually_exclusive_group()
        admin_group.add_argument(
            '--enable',
            action='store_true',
            default=None,
            help="Enable member (default)."
        )
        admin_group.add_argument(
            '--disable',
            action='store_true',
            default=None,
            help="Disable member."
        )
        parser.add_argument(
            '--wait',
            action='store_true',
            help='Wait for action to complete.',
        )
        parser.add_argument(
            '--request-sriov',
            action='store_true',
            default=None,
            help='Request that the member port be created using an SR-IOV VF.',
        )

        _tag.add_tag_option_to_parser_for_create(
            parser, 'member')

        return parser

    def take_action(self, parsed_args):
        rows = const.MEMBER_ROWS
        attrs = v2_utils.get_member_attrs(self.app.client_manager, parsed_args)

        validate.check_member_attrs(attrs)

        pool_id = attrs.pop('pool_id')

        body = {"member": attrs}
        data = self.app.client_manager.load_balancer.member_create(
            pool_id=pool_id,
            json=body
        )

        if parsed_args.wait:
            pool = self.app.client_manager.load_balancer.pool_show(pool_id)
            v2_utils.wait_for_active(
                status_f=(self.app.client_manager.load_balancer.
                          load_balancer_show),
                res_id=pool['loadbalancers'][0]['id']
            )
            data = {
                'member': (
                    self.app.client_manager.load_balancer.member_show(
                        pool_id, data['member']['id']))
            }

        # Handle older API versions that did not have the vnic_type
        if not data['member'].get('vnic_type', False):
            data['member']['vnic_type'] = 'normal'

        formatters = {'tags': v2_utils.FlatListColumn}

        return (rows,
                (utils.get_dict_properties(
                    data['member'], rows, formatters=formatters)))


class SetMember(command.Command):
    """Update a member"""

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)

        parser.add_argument(
            'pool',
            metavar='<pool>',
            help="Pool that the member to update belongs to (name or ID)."
        )
        parser.add_argument(
            'member',
            metavar='<member>',
            help="Name or ID of the member to update."
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            help="Set the name of the member."
        )
        backup = parser.add_mutually_exclusive_group()
        backup.add_argument(
            '--disable-backup',
            action='store_true',
            default=None,
            help="Disable member backup (default)."
        )
        backup.add_argument(
            '--enable-backup',
            action='store_true',
            default=None,
            help="Enable member backup."
        )
        parser.add_argument(
            '--weight',
            metavar='<weight>',
            type=int,
            help="Set the weight of member in the pool."
        )
        parser.add_argument(
            '--monitor-port',
            metavar='<monitor_port>',
            type=int,
            help="An alternate protocol port used for health monitoring a "
                 "backend member.",
        )
        parser.add_argument(
            '--monitor-address',
            metavar='<monitor_address>',
            help="An alternate IP address used for health monitoring a "
                 "backend member."
        )
        admin_group = parser.add_mutually_exclusive_group()
        admin_group.add_argument(
            '--enable',
            action='store_true',
            default=None,
            help="Set the admin_state_up to True."
        )
        admin_group.add_argument(
            '--disable',
            action='store_true',
            default=None,
            help="Set the admin_state_up to False.")
        parser.add_argument(
            '--wait',
            action='store_true',
            help='Wait for action to complete.',
        )

        _tag.add_tag_option_to_parser_for_set(parser, 'member')

        return parser

    def take_action(self, parsed_args):
        attrs = v2_utils.get_member_attrs(self.app.client_manager, parsed_args)

        validate.check_member_attrs(attrs)

        pool_id = attrs.pop('pool_id')
        member_id = attrs.pop('member_id')

        member_show = functools.partial(
            self.app.client_manager.load_balancer.member_show,
            pool_id
        )
        v2_utils.set_tags_for_set(
            member_show, member_id, attrs, clear_tags=parsed_args.no_tag)

        post_data = {"member": attrs}

        self.app.client_manager.load_balancer.member_set(
            pool_id=pool_id,
            member_id=member_id,
            json=post_data
        )

        if parsed_args.wait:
            v2_utils.wait_for_active(
                status_f=member_show,
                res_id=member_id
            )


class DeleteMember(command.Command):
    """Delete a member from a pool """

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)

        parser.add_argument(
            'pool',
            metavar='<pool>',
            help="Pool name or ID to delete the member from."
        )

        parser.add_argument(
            'member',
            metavar='<member>',
            help="Name or ID of the member to be deleted."
        )
        parser.add_argument(
            '--wait',
            action='store_true',
            help='Wait for action to complete.',
        )

        return parser

    def take_action(self, parsed_args):
        attrs = v2_utils.get_member_attrs(self.app.client_manager, parsed_args)
        id = attrs.pop('member_id')
        pool_id = attrs.pop('pool_id')

        self.app.client_manager.load_balancer.member_delete(
            pool_id=pool_id,
            member_id=id
        )

        if parsed_args.wait:
            member_show = functools.partial(
                self.app.client_manager.load_balancer.member_show,
                pool_id
            )
            v2_utils.wait_for_delete(
                status_f=member_show,
                res_id=id
            )


class UnsetMember(command.Command):
    """Clear member settings"""

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)

        parser.add_argument(
            'pool',
            metavar='<pool>',
            help="Pool that the member to update belongs to (name or ID)."
        )
        parser.add_argument(
            'member',
            metavar="<member>",
            help="Member to modify (name or ID)."
        )
        parser.add_argument(
            '--backup',
            action='store_true',
            help="Clear the backup member flag."
        )
        parser.add_argument(
            '--monitor-address',
            action='store_true',
            help="Clear the member monitor address."
        )
        parser.add_argument(
            '--monitor-port',
            action='store_true',
            help="Clear the member monitor port."
        )
        parser.add_argument(
            '--name',
            action='store_true',
            help="Clear the member name."
        )
        parser.add_argument(
            '--weight',
            action='store_true',
            help="Reset the member weight to the API default."
        )
        parser.add_argument(
            '--wait',
            action='store_true',
            help='Wait for action to complete.',
        )
        _tag.add_tag_option_to_parser_for_unset(parser, 'member')

        return parser

    def take_action(self, parsed_args):
        unset_args = v2_utils.get_unsets(parsed_args)
        if not unset_args and not parsed_args.all_tag:
            return

        pool_id = v2_utils.get_resource_id(
            self.app.client_manager.load_balancer.pool_list,
            'pools', parsed_args.pool)

        member_show = functools.partial(
            self.app.client_manager.load_balancer.member_show,
            pool_id
        )

        member_dict = {'pool_id': pool_id, 'member_id': parsed_args.member}
        member_id = v2_utils.get_resource_id(
            self.app.client_manager.load_balancer.member_list,
            'members', member_dict)

        v2_utils.set_tags_for_unset(
            member_show, member_id, unset_args,
            clear_tags=parsed_args.all_tag)

        body = {'member': unset_args}

        self.app.client_manager.load_balancer.member_set(
            pool_id=pool_id, member_id=member_id, json=body)

        if parsed_args.wait:
            v2_utils.wait_for_active(
                status_f=member_show,
                res_id=member_id
            )



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from osc_lib import exceptions

from octaviaclient.osc.v2 import constants


def check_l7policy_attrs(attrs):
    msg = None
    if 'action' not in attrs:
        return
    if attrs['action'] == 'REDIRECT_TO_POOL':
        if 'redirect_pool_id' not in attrs:
            msg = 'Missing argument: --redirect-pool'
    elif attrs['action'] == 'REDIRECT_TO_URL':
        if 'redirect_url' not in attrs:
            msg = 'Missing argument: --redirect-url'
    elif attrs['action'] == 'REDIRECT_PREFIX':
        if 'redirect_prefix' not in attrs:
            msg = 'Missing argument: --redirect-prefix'
    if msg is not None:
        raise exceptions.CommandError(msg)


def check_l7rule_attrs(attrs):
    if 'type' in attrs:
        if attrs['type'] in ('COOKIE', 'HEADER'):
            if 'key' not in attrs:
                msg = (
                    "Missing argument: --type {type_name} requires "
                    "--key <key>".format(type_name=attrs['type']))
                raise exceptions.CommandError(msg)


# Handling these range validations here instead of "choices" as "choices" will
# output every possible option in the error message.
def _validate_TCP_UDP_SCTP_port_range(port_number, parameter_name):
    if (port_number < constants.MIN_PORT_NUMBER or
            port_number > constants.MAX_PORT_NUMBER):
        msg = ("Invalid input for field/attribute '{name}', Value: "
               "'{port}'. Value must be between {pmin} and {pmax}.".format(
                   name=parameter_name, port=port_number,
                   pmin=constants.MIN_PORT_NUMBER,
                   pmax=constants.MAX_PORT_NUMBER))
        raise exceptions.InvalidValue(msg)


def check_listener_attrs(attrs):
    if 'protocol_port' in attrs:
        _validate_TCP_UDP_SCTP_port_range(attrs['protocol_port'],
                                          'protocol-port')

    extra_hsts_opts_set = attrs.get('hsts_preload') or attrs.get(
        'hsts_include_subdomains')
    if extra_hsts_opts_set and 'hsts_max_age' not in attrs:
        raise exceptions.InvalidValue(
            "Argument hsts_max_age is required when using hsts_preload or "
            "hsts_include_subdomains arguments.")


def check_member_attrs(attrs):
    if 'protocol_port' in attrs:
        _validate_TCP_UDP_SCTP_port_range(attrs['protocol_port'],
                                          'protocol-port')

    if 'member_port' in attrs:
        _validate_TCP_UDP_SCTP_port_range(attrs['member_port'], 'member-port')

    if 'weight' in attrs:
        if (attrs['weight'] < constants.MIN_WEIGHT or
                attrs['weight'] > constants.MAX_WEIGHT):
            msg = ("Invalid input for field/attribute 'weight', Value: "
                   "'{weight}'. Value must be between {wmin} and "
                   "{wmax}.".format(weight=attrs['weight'],
                                    wmin=constants.MIN_WEIGHT,
                                    wmax=constants.MAX_WEIGHT))
            raise exceptions.InvalidValue(msg)
